from .drivers import current_driver, Driver
from .data_variable import DataVariable
from .support_data_variable import SupportDataVariable
import re
import numpy as np
from typing import List, Optional
import logging

DEPEND_REGEX = re.compile("DEPEND_\\d")

ISTP_NOT_COMPLIANT_W = "Non compliant ISTP file"

log = logging.getLogger(__name__)


def _get_attributes(master_cdf: Driver, cdf: Driver, var: str):
    attrs = {}
    for attr in master_cdf.variable_attributes(var):
        value = master_cdf.variable_attribute_value(var, attr)
        if attr.endswith("_PTR") or attr[:-1].endswith("_PTR_"):
            if master_cdf.has_variable(value):
                value = master_cdf.values(value, is_metadata_variable=True)
                if hasattr(value, 'tolist'):
                    attrs[attr] = value.tolist()
                else:
                    attrs[attr] = value
            else:
                log.warning(
                    f"{ISTP_NOT_COMPLIANT_W}: variable {var} has {attr} attribute which points to variable {value} which does not exist")
                attrs[attr] = value
        else:
            attrs[attr] = value
    return attrs


def _get_axis(master_cdf: Driver, cdf: Driver, axis_var: str, data_var: str):
    src_cdf = cdf if cdf.has_variable(axis_var) else master_cdf if master_cdf.has_variable(axis_var) else None
    if src_cdf is not None:
        if src_cdf.is_char(axis_var):
            if 'sig_digits' in master_cdf.variable_attributes(axis_var):  # cluster CSA trick :/
                return SupportDataVariable(name=axis_var, values=np.asarray(src_cdf.values(axis_var), dtype=float),
                                           attributes=_get_attributes(master_cdf, src_cdf, axis_var),
                                           is_nrv=src_cdf.is_nrv(axis_var),
                                           cdf_type=src_cdf.cdf_type(axis_var)
                                           )
        return SupportDataVariable(name=axis_var, values=src_cdf.values(axis_var),
                                   attributes=_get_attributes(master_cdf, src_cdf, axis_var),
                                   is_nrv=src_cdf.is_nrv(axis_var),
                                   cdf_type=src_cdf.cdf_type(axis_var)
                                   )
    else:
        log.warning(
            f"{ISTP_NOT_COMPLIANT_W}: trying to load {axis_var} as support data for {data_var} but it is absent from the file")
    return None


def _get_axes(master_cdf: Driver, cdf: Driver, var: str, data_shape):
    attrs = sorted(filter(lambda attr: DEPEND_REGEX.match(attr), master_cdf.variable_attributes(var)))
    unix_time_name = master_cdf.variable_attribute_value(var, "DEPEND_TIME")
    axes = list(
        map(lambda attr: _get_axis(master_cdf, cdf, master_cdf.variable_attribute_value(var, attr), var), attrs))
    if unix_time_name is not None and unix_time_name in master_cdf.variables():
        unix_time = _get_axis(master_cdf, cdf, unix_time_name, var)
        if len(unix_time) == data_shape[0] and len(axes[0].values) != data_shape[0]:
            unix_time.values = (unix_time.values * 1e9).astype('<M8[ns]')
            axes[0] = unix_time
            log.warning(
                f"{ISTP_NOT_COMPLIANT_W}: swapping DEPEND_0 with DEPEND_TIME for {var}, if you think this is a bug report it here: https://github.com/SciQLop/PyISTP/issues")
    return axes


def _get_labels(attributes) -> List[str]:
    if 'LABL_PTR_1' in attributes:
        return attributes['LABL_PTR_1']
    if 'LABLAXIS' in attributes:
        return [attributes['LABLAXIS']]


def _load_data_var(master_cdf: Driver, cdf: Driver, var: str) -> DataVariable or None:
    values = lambda: cdf.values(var)
    shape = cdf.shape(var)
    axes = _get_axes(master_cdf, cdf, var, shape)
    attributes = _get_attributes(master_cdf, cdf, var)
    cdf_type = cdf.cdf_type(var)
    labels = _get_labels(attributes)
    if len(axes) == 0:
        log.warning(f"{ISTP_NOT_COMPLIANT_W}: {var} was marked as data variable but it has 0 support variable")
        return None
    if None in axes or axes[0].values.shape[0] != shape[0]:
        return None
    return DataVariable(name=var, values=values, attributes=attributes, axes=axes, cdf_type=cdf_type, labels=labels)


class ISTPLoaderImpl:
    cdf: Optional[Driver] = None

    def __init__(self, file=None, buffer=None, master_file=None, master_buffer=None):
        if file is not None:
            log.debug(f"Loading {file}")
        self.cdf = current_driver(file or buffer)
        if master_file or master_buffer:
            self.master_cdf = current_driver(master_file or master_buffer)
        else:
            self.master_cdf = self.cdf
        self.data_variables = []
        self._update_data_vars_lis()

    def attributes(self):
        return self.master_cdf.attributes()

    def attribute(self, key):
        return self.master_cdf.attribute(key)

    def _update_data_vars_lis(self):
        if self.master_cdf:
            self.data_variables = []
            for var in self.master_cdf.variables():
                var_attrs = self.master_cdf.variable_attributes(var)
                var_type = self.master_cdf.variable_attribute_value(var, 'VAR_TYPE')
                param_type = (self.master_cdf.variable_attribute_value(var,
                                                                       'PARAMETER_TYPE') or "").lower()  # another cluster CSA crap
                if (var_type == 'data' or param_type == 'data') and not self.master_cdf.is_char(var):
                    self.data_variables.append(var)
            if len(self.data_variables) == 0:
                log.warning(f"{ISTP_NOT_COMPLIANT_W}: No data variable found, this is suspicious")

    def data_variable(self, var_name) -> DataVariable:
        return _load_data_var(self.master_cdf, self.cdf, var_name)



import pycdfpp
import numpy as np


def _drop_first_dim_if_nrv(is_nrv: bool, values):
    if is_nrv:
        if values.shape[0] == 1:
            return values[0]
        else:
            return np.array([], dtype=values.dtype)
    else:
        return values


class Driver:
    def __init__(self, file):
        self.cdf = pycdfpp.load(file)

    def attributes(self):
        if self.cdf:
            return self.cdf.attributes.keys()
        return []

    def attribute(self, key):
        if self.cdf:
            return self.cdf.attributes[key]
        return None

    def variables(self):
        if self.cdf:
            return list(self.cdf)
        return []

    def cdf_type(self, var_name):
        return self.cdf[var_name].type.name

    def has_variable(self, name):
        return name in self.cdf

    def is_char(self, var):
        return self.cdf[var].type == pycdfpp.DataType.CDF_CHAR

    def variable_attributes(self, var):
        if self.cdf:
            return self.cdf[var].attributes.keys()
        return []

    def variable_attribute_value(self, var, attr):
        if self.cdf and var in self.cdf and attr in self.cdf[var].attributes:
            return self.cdf[var].attributes[attr][0]
        return None

    def values(self, var, is_metadata_variable=False):
        v: pycdfpp.Variable = self.cdf[var]
        if v.type in (pycdfpp.DataType.CDF_EPOCH, pycdfpp.DataType.CDF_EPOCH16, pycdfpp.DataType.CDF_TIME_TT2000):
            return _drop_first_dim_if_nrv(v.is_nrv, pycdfpp.to_datetime64(v))
        if is_metadata_variable and self.is_char(var):
            return _drop_first_dim_if_nrv(v.is_nrv, v.values_encoded)
        return _drop_first_dim_if_nrv(v.is_nrv, v.values)

    def is_nrv(self, var):
        return self.cdf[var].is_nrv

    def shape(self, var):
        return self.cdf[var].shape



"""Top-level package for Python ISTP loader."""

__author__ = """Alexis Jeandet"""
__email__ = 'alexis.jeandet@member.fsf.org'
__version__ = '0.7.2'

from .loader import ISTPLoader as _ISTPLoader


def load(file=None, buffer=None, master_file=None, master_buffer=None) -> _ISTPLoader:
    return _ISTPLoader(file=file, buffer=buffer, master_file=master_file, master_buffer=master_buffer)



from typing import Protocol, runtime_checkable, List, AnyStr, Collection, Any


class Driver(Protocol):

    def attributes(self) -> List[AnyStr]:
        ...

    def attribute(self, key: AnyStr):
        ...

    def data_variables(self) -> List[AnyStr]:
        ...

    def data_variable(self, var_name: AnyStr):
        ...

    def cdf_type(self, var_name: AnyStr) -> AnyStr:
        ...

    def variable_attribute_value(self, var: AnyStr, attr: AnyStr) -> Any:
        ...

    def variable_attributes(self, var: AnyStr) -> List[AnyStr]:
        ...

    def values(self, var: AnyStr, is_metadata_variable: bool = False) -> Collection:
        ...

    def shape(self, var: AnyStr) -> Collection[int]:
        ...

    def variables(self) -> List[AnyStr]:
        ...

    def has_variable(self, name: str) -> bool:
        ...

    def is_char(self, var: str) -> bool:
        ...

    def is_nrv(self, var: str) -> bool:
        ...



from typing import Collection, Callable, Any, Union


class DataVariable:
    __slots__ = ("name", "_values", 'attributes', 'axes', 'labels', 'cdf_type')

    def __init__(self, name: str, values: Union[Collection[Any], Callable], attributes, axes, cdf_type, labels=None):
        self.name = name
        self._values = values
        self.attributes = attributes
        self.axes = axes or []
        self.labels = labels
        self.cdf_type = cdf_type

    @property
    def values(self):
        if callable(self._values):
            self._values = self._values()
        return self._values

    def __len__(self):
        return len(self.values)

    def __repr__(self):
        return f"""DataVariable: {self.name}
Attributes:
{self.attributes}
Axes:
{self.axes}\n"""



import numpy as np


class SupportDataVariable:
    __slots__ = ("name", "values", 'attributes', 'is_nrv', 'cdf_type')

    def __init__(self, name, values, attributes, is_nrv, cdf_type):
        self.name = name
        self.values = values
        if type(self.values) is list:
            self.values = np.array(self.values)
        self.attributes = attributes
        self.is_nrv = is_nrv
        self.cdf_type = cdf_type

    def __len__(self):
        return len(self.values)

    def __repr__(self):
        return f"""SupportDataVariable: {self.name}
Attributes:
{self.attributes}\n"""
